#include "ProtoScript.h"

#include "chuck_fft.h"
#include "MusicCommons.h"

#include "ProtoFault.h"
#include "ProtoSmooth.h"
#include "ProtoBlend.h"
#include "ProtoSong.h"

#include <cstdio>
#include <cassert>
#include <cmath>
#include <iostream>

using namespace std;

#define SUBBAND_DIV ( NSUBBANDS/2 )

ProtoScript::ProtoScript()
{
	hanning( window_, window_size_ );

	for( uint32_t sb = 0; sb < NSUBBANDS; sb++ ) {
		subband_detector_[sb] = new BeatDetector( 44100 / window_size_ );
	}

	events_ = new ProtoEventQueue();
}

ProtoScript::~ProtoScript()
{
	for( uint32_t sb = 0; sb < NSUBBANDS; sb++ ) {
		delete subband_detector_[sb];
	}

	while( MoreEvents() ) {
		delete NextEvent();
	}
}

// Get a viable percentage number we can use to index into the terrain
float normalize( float num )
{
	float norm = fabs( num );

	if( norm > 1.1f ) {
		do {
			norm /= 10.0f;
		} while( norm > 1.1f );
	}

	if( norm < 0.1f ) {
		do {
			norm *= 10.0f;
		} while( norm < 0.1f );
	}

	norm -= 0.1f;

	assert( 0 <= norm && 1 >= norm );

	return norm;
}

void ProtoScript::HandleInstant(
		StereoPCM const _playlist[], unsigned long _song,  uint64_t _bookmark )
{
	RT_SAMPLE instant[window_size_*2];
	StereoPCM const *pcm = &_playlist[_song];

	for( uint32_t i = 0; i < 2 * window_size_; i += 2) {
		uint32_t left = i;
		uint32_t right = i+1;

		if( pcm->left->size() > i + _bookmark ) {
			// Convert to [-1,1] float for fft (is this necessary?)
			instant[left] = (RT_SAMPLE)pcm->left->at( i + _bookmark ) / 32768.0f;
			instant[right] = (RT_SAMPLE)pcm->right->at( i + _bookmark ) / 32768.0f;
		} else {
			instant[left] = instant[right] = 0;
		}
	}

	// Based on algorithms at:
	// http://www.gamedev.net/reference/programming/features/beatdetection/

	RT_SAMPLE fft_mono[window_size_];

	for( uint32_t i = 0; i < window_size_; i++ ) {
		fft_mono[i] = ( instant[2*i] + instant[2*i+1] ) / 2.0f;
	}

	apply_window( fft_mono, window_, window_size_ );

	rfft( (RT_SAMPLE *)fft_mono, window_size_/2, FFT_FORWARD );

	// Now, do the above, but with fft'd value
	// Instead of left & right channels, use real and complex components

	// for each subband...
	uint32_t start = 0;
	for( uint32_t sb = 0; sb < NSUBBANDS; sb++ ) {
		complex *cbuf = (complex *)fft_mono;
		complex sb_caverage;
		RT_SAMPLE sb_instant_energy_sum = 0;

		sb_caverage.re = 0;
		sb_caverage.im = 0;

		// for each sample in that subband...
		for( uint32_t i = start; i < start + g_subband_width[sb]; i++ ) {
			sb_instant_energy_sum += cmp_abs( cbuf[i] );
			sb_caverage.re += cbuf[i].re;
			sb_caverage.im += cbuf[i].im;
		}
		start += g_subband_width[sb];

		sb_caverage.re /= g_subband_width[sb];
		sb_caverage.im /= g_subband_width[sb];

		RT_SAMPLE sb_instant_energy_average =
			sb_instant_energy_sum / g_subband_width[sb];

		subband_detector_[sb]->tick( sb_instant_energy_average );

		// TODO This is in desperate need of refactoring
		if( subband_detector_[sb]->is_beat() ) {
			ProtoEvent *newEvent = NULL;

			if( sb == 4 || sb == 5 ) { 
				newEvent =
					new ProtoFault(
							normalize( sb_caverage.re), normalize( sb_caverage.im ),
							( SUBBAND_DIV - sb ) * 5 );
			} else if( sb == SUBBAND_DIV ) {
				newEvent =
					new ProtoSmooth(
							(float)(sb - (SUBBAND_DIV-1)) /
							(float)(NSUBBANDS - (SUBBAND_DIV-1)) );
			}

			if( newEvent ) {
				newEvent->indexIs( _song );
				newEvent->startTimeIs( _bookmark * 1000 / 44100 ); /* ms since start */
				newEvent->endTimeIs( newEvent->startTime() ); /* not currently used */
				events_->push( newEvent );
			}
		}
	}
}

void ProtoScript::Freeze( string _filename, StereoPCM const _playlist[] )
{
	ProtoScript *script = new ProtoScript();

	FILE *to_file = fopen( _filename.c_str(), "w" );

	for( uint32_t song = 0; song < N_SONGS; song++ ) {
		ProtoEvent *newEvent = new ProtoSong( _playlist[song].filename );
		newEvent->startTimeIs( 0 ); // All songs start from beginning
		newEvent->indexIs( song );

		for( uint32_t smp = 0; smp < _playlist[song].left->size();
				smp += window_size_ ) {
			script->HandleInstant( _playlist, song, smp );
		}

		newEvent->endTimeIs( 0 ); // TODO Unused, but could be used to indicate len

		script->events_->push( newEvent );
	}

	while( script->MoreEvents() ) {
		ProtoEvent *event = script->NextEvent();

		event->Freeze( to_file );

		delete event;
	}

	fclose( to_file );

	delete script;
}

ProtoScript *ProtoScript::UnFreeze(	string _filename )
{
	ProtoScript *script = new ProtoScript();

	FILE *from_file = fopen( _filename.c_str(), "r" );

	assert( from_file );

	char buf[2048];

	while( fgets( buf, 2047, from_file ) ) {
		ProtoEvent *newEvent = NULL;
		string str( buf );

		if( str.find( "Fault" ) != string::npos ) {
			newEvent = new ProtoFault();
		} else if( str.find( "Smooth" ) != string::npos ) {
			newEvent = new ProtoSmooth();
		} else if( str.find( "Blend" ) != string::npos ) {
			newEvent = new ProtoBlend();
		} else if( str.find( "Song" ) != string::npos ) {
			newEvent = new ProtoSong();
		} else {
			cerr << str << endl;
		}

		if( newEvent ) {
			newEvent->UnFreeze( str );
			script->events_->push( newEvent );
		}
	}

	fclose( from_file );

	return script;
}

bool ProtoScript::MoreEvents()
{
	return !events_->empty();
}

ProtoEvent *ProtoScript::NextEvent()
{
	ProtoEvent *event = events_->top();
	events_->pop();
	return event;
}
