#ifndef __UNITS_H__
#define __UNITS_H__

#include <stdlib.h>
#include <math.h>

#define MY_PIE 3.14159265358979

// ------------
// Classes
//   Nominal and Ordinal are based heavily on Prof. David Cheriton's framework code
//   from CS249A and CS249B.
// ------------
template <class UNIT_TYPE,class REP_TYPE>
class Nominal {
	public:
		Nominal() {}
		Nominal(REP_TYPE newRep) : rep(newRep) {}

		REP_TYPE value() const {
			return rep;
		}

		bool operator==(const Nominal<UNIT_TYPE,REP_TYPE> &rhs) const {
			return rhs.rep == rep;
		}

		bool operator!=(const Nominal<UNIT_TYPE,REP_TYPE> &rhs) const {
			return rhs.rep != rep;
		}

		const Nominal<UNIT_TYPE,REP_TYPE> operator=(const Nominal<UNIT_TYPE,REP_TYPE> &rhs) {
			rep = rhs.rep;
			return *this;
		}

		const Nominal<UNIT_TYPE,REP_TYPE> operator=(const REP_TYPE &rhs) {
			rep = rhs;
			return *this;
		}

	protected:
		REP_TYPE rep;
};

template <class UNIT_TYPE,class REP_TYPE>
class Ordinal : public Nominal<UNIT_TYPE,REP_TYPE> {
	// TODO add operators
	public:
		Ordinal() {}
		Ordinal(REP_TYPE newRep) : Nominal<UNIT_TYPE,REP_TYPE>(newRep) {}

		Ordinal<UNIT_TYPE,REP_TYPE> operator+(const Ordinal<UNIT_TYPE,REP_TYPE> &rhs) {
			return Nominal<UNIT_TYPE,REP_TYPE>::rep + rhs.rep;
		}

		Ordinal<UNIT_TYPE,REP_TYPE> operator+=(const Ordinal<UNIT_TYPE,REP_TYPE> &rhs) {
			Nominal<UNIT_TYPE,REP_TYPE>::rep += rhs.rep;
			return *this;
		}

		Ordinal<UNIT_TYPE,REP_TYPE> operator++(int) {
			REP_TYPE rv = Nominal<UNIT_TYPE,REP_TYPE>::rep;
			Nominal<UNIT_TYPE,REP_TYPE>::rep++;
			return rv;
		}
};

class Frequency : public Ordinal<Frequency,int>{
	public:
		static const int NO_FREQUENCY = -1;

		Frequency() : Ordinal<Frequency,int>(NO_FREQUENCY) {}
		Frequency(int newRep) : Ordinal<Frequency,int>(newRep) {}

		Frequency operator=(const std::string &rhs) {
			// TODO replace atof with strtod for error checking
			rep = atof(rhs.c_str());
			return *this;
		}

		Frequency operator=(const char *rhs) {
			rep = atof(rhs);
			return *this;
		}
};

class Width : public Ordinal<Width,double> {
	public:
		static const double NO_WIDTH = -1.0;

		Width() : Ordinal<Width,double>(NO_WIDTH) {}

		Width operator=(const std::string &rhs) {
			rep = atof(rhs.c_str());
			return *this;
		}

		Width operator=(const char *rhs) {
			rep = atof(rhs);
			return *this;
		}
};

class SignalFrame : public Ordinal<SignalFrame,double> {
	public:
		const static double STARTING_FRAME = 0.0;

		SignalFrame() : Ordinal<SignalFrame,double>(STARTING_FRAME) {}
		SignalFrame(double newRep) : Ordinal<SignalFrame,double>(newRep) {}
};

class Seconds : public Ordinal<Seconds,double> {
	public:
		const static double STARTING_TIME = 0.0;

		Seconds() : Ordinal<Seconds,double>(STARTING_TIME) {}
};

class Amplitude : public Ordinal<Amplitude,double> {
	public:
		const static double NO_AMPLITUDE = 0;

		Amplitude() : Ordinal<Amplitude,double>(NO_AMPLITUDE) {}
		Amplitude(double newRep) : Ordinal<Amplitude,double>(newRep) {}

		static Amplitude random() {
			return Amplitude(((double)rand() - (double)RAND_MAX / 2.0) / ((double)RAND_MAX / 2.0));
		}
};

class Sample : public Ordinal<Sample,int> {
	public:
		static const double FIRST_SAMPLE = 0;

		Sample() : Ordinal<Sample,int>(FIRST_SAMPLE) {}
		Sample(int newRep) : Ordinal<Sample,int>(newRep) {}
};

#define KILLCHECK(condition,msg) (condition) ? std::cout << msg << std::endl, exit(-1) : (void)0

// TODO Range checks belong in value-type classes
#define WIDTH_RANGE() KILLCHECK((width.value() < 0.0 || width.value() > 1.0), "Width must be between 0 and 1.")
#define FREQUENCY_RANGE() KILLCHECK((freq.value() < 0.0 || freq.value() > 20000.0), "Frequency must be between 0 and 20000.")

// TODO Require/forbid checks belong in to-be-written value-type classes
#define REQUIRE_FREQUENCY(msg) KILLCHECK((freq.value() == Frequency::NO_FREQUENCY),msg), FREQUENCY_RANGE()
#define FORBID_FREQUENCY(msg) KILLCHECK((freq.value() != Frequency::NO_FREQUENCY),msg)
#define REQUIRE_WIDTH(msg) KILLCHECK((width.value() == Width::NO_WIDTH), msg), WIDTH_RANGE()
#define FORBID_WIDTH(msg) KILLCHECK((width.value() != Width::NO_WIDTH), msg)

class Signal {
	public:
		virtual void fillBuffer(double *buffer, unsigned int nFrames) = 0;
		virtual void assertValid() = 0;

	protected:
		friend class SignalFactory;
		Signal(unsigned int _nChannels = 2) : nChannels(_nChannels) {}
		Signal(Frequency _sampleRate, Frequency _freq, Width _width, unsigned int _nChannels = 2) : sampleRate(_sampleRate), freq(_freq), width(_width), nChannels(_nChannels) {}

		Frequency sampleRate;
		Frequency freq;
		Width width;
		Sample sample;
		unsigned int nChannels;
};

class Sine : public Signal {
	public:
		virtual void fillBuffer(double *buffer, unsigned int nFrames) {
			for( unsigned int i = 0; i < nChannels * nFrames; i += nChannels ) {
				sample++;
				double ft = freq.value() * sample.value() / sampleRate.value();

				for( unsigned int j = 0; j < nChannels; j++ ) {
					buffer[i+j] = sin( 2 * MY_PIE * ft );
				}
			}
		}

		virtual void assertValid() {
			REQUIRE_FREQUENCY("Sine wave requires a frequency.");
			FORBID_WIDTH("Width makes no sense with a sine wave.");
		}
	protected:
		friend class SignalFactory;
		Sine(Frequency _sampleRate, Frequency _freq, Width _width, unsigned int _nChannels) : Signal(_sampleRate, _freq, _width, _nChannels) {}
};

class Saw : public Signal {
	public:
		virtual void fillBuffer(double *buffer, unsigned int nFrames) {
			int samplesPerPeriod = (int)(sampleRate.value() / freq.value());
			int peakAtSample = (int)(samplesPerPeriod * width.value());

			// HACKHACK
			if( peakAtSample == 0) {
				peakAtSample = 1;
			}

			for( unsigned int i = 0; i < nChannels * nFrames; i += nChannels ) {
				sample++;

				int samplesIntoPeriod = sample.value() % samplesPerPeriod;

				for( unsigned int j = 0; j < nChannels; j++ ) {
					if( samplesIntoPeriod < peakAtSample) {
						buffer[i+j] = -1.0 + 2 * (double)samplesIntoPeriod / peakAtSample;
					} else {
						buffer[i+j] = 1.0 - 2 * (double)(samplesIntoPeriod - peakAtSample) / (samplesPerPeriod - peakAtSample);
					}
				}
			}
		}

		virtual void assertValid() {
			REQUIRE_FREQUENCY("Saw requires a frequency.");
			REQUIRE_WIDTH("Saw requires a width.");
		}
	protected:
		friend class SignalFactory;
		Saw(Frequency _sampleRate, Frequency _freq, Width _width, unsigned int _nChannels) : Signal(_sampleRate, _freq, _width, _nChannels) {}
};

class Pulse : public Signal {
	public:
		virtual void fillBuffer(double *buffer, unsigned int nFrames) {
			int samplesPerPeriod = (int)(sampleRate.value() / freq.value());
			int transitionAtSample = (int)(samplesPerPeriod * width.value());

			for( unsigned int i = 0; i < nChannels * nFrames; i += nChannels ) {
				sample++;

				int samplesIntoPeriod = sample.value() % samplesPerPeriod;

				for( unsigned int j = 0; j < nChannels; j++ ) {
					buffer[i+j] = samplesIntoPeriod < transitionAtSample ? 1.0 : -1.0;
				}
			}
		}

		virtual void assertValid() {
			REQUIRE_FREQUENCY("Pulse requires a frequency.");
			REQUIRE_WIDTH("Pulse requires a width.");
		}
	protected:
		friend class SignalFactory;
		Pulse(Frequency _sampleRate, Frequency _freq, Width _width, unsigned int _nChannels) : Signal(_sampleRate, _freq, _width, _nChannels) {}
};

class Noise : public Signal {
	public:
		virtual void fillBuffer(double *buffer, unsigned int nFrames) {
			for( unsigned int i = 0; i < nChannels * nFrames; i += nChannels ) {
				sample++;
				for( unsigned int j = 0; j < nChannels; j++ ) {
					buffer[i+j] = Amplitude::random().value();
				}
			}
		}

		virtual void assertValid() {
			FORBID_FREQUENCY("Frequency makes no sense for noise.");
			FORBID_WIDTH("Width makes no sense for noise.");
		}
	protected:
		friend class SignalFactory;
		Noise(Frequency _sampleRate, Frequency _freq, Width _width, unsigned int _nChannels) : Signal(_sampleRate, _freq, _width, _nChannels) {}
};

class Impulse : public Signal {
	public:
		virtual void fillBuffer(double *buffer, unsigned int nFrames) {
			int samplesPerPeriod = (int)(sampleRate.value() / freq.value());

			for( unsigned int i = 0; i < nChannels * nFrames; i += nChannels ) {
				sample++;

				int samplesIntoPeriod = sample.value() % samplesPerPeriod;

				for( unsigned int j = 0; j < nChannels; j++  ) {
					buffer[i+j] = sample.value() % (samplesIntoPeriod+1) == 0 ? 1.0 : 0.0;
				}
			}
		}

		virtual void assertValid() {
			REQUIRE_FREQUENCY("Impulse requires a frequency.");
			FORBID_WIDTH("Width makes no sense for impulse.");
		}
	protected:
		friend class SignalFactory;
		Impulse(Frequency _sampleRate, Frequency _freq, Width _width, unsigned int _nChannels) : Signal(_sampleRate, _freq, _width, _nChannels) {}
};

class SignalFactory {
	public:
		static Signal *newSignal(std::string type, Frequency sampleRate, Frequency freq = Frequency(), Width width = Width(), unsigned int nChannels = 2) {
			std::string newTypeCopy(type);

			for( unsigned int i = 0; i < type.size(); i++ ) {
				newTypeCopy[i] = tolower(newTypeCopy[i]);
			}

			if( newTypeCopy == "sine" ) {
				return new Sine(sampleRate, freq, width, nChannels);
			} else if( newTypeCopy == "saw" ) {
				return new Saw(sampleRate, freq, width, nChannels);
			} else if( newTypeCopy == "pulse" ) {
				return new Pulse(sampleRate, freq, width, nChannels);
			} else if( newTypeCopy == "noise" ) {
				return new Noise(sampleRate, freq, width, nChannels);
			} else if( newTypeCopy == "impulse" ) {
				return new Impulse(sampleRate, freq, width, nChannels);
			}

			return NULL; /* Invalid type */
		}
};

#endif
