#include "RtAudio.h"
#include "Units.h"
#include "mad.h"
#include "chuck_fft.h"

#include "MusicCommons.h"
#include "Mutex.h"
#include "Mad.h"
#include "BeatDetector.h"

#include "ProtoScript.h"

#include <unistd.h>
#include <stdint.h>
#include <sys/time.h>
#include <fcntl.h>
#include <pthread.h>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <cstring>
#include <cassert>

#include <deque>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

using namespace std;

//-----------------------------------------------------------------------------
// function prototypes
//-----------------------------------------------------------------------------
void idleFunc( );
void displayFunc( );
void reshapeFunc( GLsizei width, GLsizei height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );

// width and height of the window
GLsizei g_width = 800;
GLsizei g_height = 600;

// ---------------------
// Globals for this file
// ---------------------
deque<MAD_SAMPLE> *g_pcm_left;
deque<MAD_SAMPLE> *g_pcm_right;

BeatDetector *g_subband_detector[NSUBBANDS];
GLfloat g_subband_color[NSUBBANDS][3];

struct timeval g_last_draw;

RT_SAMPLE *g_window;
RT_SAMPLE *g_fft_buffer;
Mutex *g_fft_buffer_lock;
uint32_t g_buffer_size;

#define STEREO 2
#define NCHANNELS STEREO

Frequency g_sampleRate;

/* RTAUDIO */

// callback
static int audioCallback( void *outputBuffer, void *inputBuffer,
		unsigned int nFrames, double streamTime,
		RtAudioStreamStatus status, void *userData )
{
	RT_SAMPLE *out = (RT_SAMPLE *)outputBuffer;
	RT_SAMPLE instant_energy_sum = 0;

	for( unsigned int i = 0; i < NCHANNELS * nFrames; i += NCHANNELS ) {
		unsigned int left = i;
		unsigned int right = i+1;

		assert( g_pcm_left->size() == g_pcm_right->size() );

		if( g_pcm_left->size() ) {
			// Convert to [-1,1] float for fft (is this necessary?)
			out[left] = (RT_SAMPLE)g_pcm_left->front() / 32768.0f;
			out[right] = (RT_SAMPLE)g_pcm_right->front() / 32768.0f;
			g_pcm_left->pop_front();
			g_pcm_right->pop_front();
		} else {
			out[left] = out[right] = 0;
		}

		RT_SAMPLE sample_energy = pow( out[left], 2 ) + pow( out[right], 2 );

		// Compute change to avoid redoing a lot of arithmetic
		instant_energy_sum += sample_energy;
	}

	// Based on algorithms at:
	// http://www.gamedev.net/reference/programming/features/beatdetection/

	if( g_window ) {
		RT_SAMPLE fft_mono[nFrames];

		for( uint32_t i = 0; i < nFrames; i++ ) {
			fft_mono[i] = ( out[2*i] + out[2*i+1] ) / 2.0f;
		}

		g_fft_buffer_lock->Lock();
		{
			memcpy( g_fft_buffer, fft_mono, nFrames * sizeof( RT_SAMPLE ) );

			// TODO Try this with and without window. We probably want it, though.
			apply_window( g_fft_buffer, g_window, nFrames );

			rfft( (RT_SAMPLE *)g_fft_buffer, nFrames/2, FFT_FORWARD );
		}
		g_fft_buffer_lock->Unlock();

		// Now, do the above, but with fft'd value
		// Instead of left & right channels, use real and complex components

		// for each subband...
		uint32_t start = 0;
		for( uint32_t sb = 0; sb < NSUBBANDS; sb++ ) {
			complex *cbuf = (complex *)g_fft_buffer;
			RT_SAMPLE sb_instant_energy_sum = 0;

			// for each sample in that subband...
			for( uint32_t i = start; i < start + g_subband_width[sb]; i++ ) {
				sb_instant_energy_sum += cmp_abs( cbuf[i] );
			}
			start += g_subband_width[sb];

			RT_SAMPLE sb_instant_energy_average =
				sb_instant_energy_sum / g_subband_width[sb];

			g_subband_detector[sb]->tick( sb_instant_energy_average );
		}
	}

	return 0;
}

/*
 * This is perhaps the simplest example use of the MAD high-level API.
 * Standard input is mapped into memory via mmap(), then the high-level API
 * is invoked with three callbacks: input, output, and error. The output
 * callback converts MAD's high-resolution PCM samples to 16 bits, then
 * writes them to standard output in little-endian, stereo-interleaved
 * format.
 */

// entry point
int main( int argc, char * argv[] )
{
	// RtAudio pointer
	RtAudio * audio = NULL;
	RtAudio::StreamParameters outParams;

	g_buffer_size = 1024;

	int fildes = -1;
	if (argc == 1) {
		fildes = STDIN_FILENO;
	} else {
		fildes = open( argv[1], O_RDONLY );
	}

	assert( fildes != -1 );

	// initialize GLUT
	glutInit( &argc, argv );
	// double buffer, use rgb color, enable depth buffer
	glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
	// initialize the window size
	glutInitWindowSize( g_width, g_height );
	// set the window postion
	glutInitWindowPosition( 100, 100 );
	// create the window
	glutCreateWindow( "Subband Beat Detection" );

	// set the idle function - called when idle
	glutIdleFunc( idleFunc );
	// set the display function - called when redrawing
	glutDisplayFunc( displayFunc );
	// set the reshape function - called when client area changes
	glutReshapeFunc( reshapeFunc );
	// set the keyboard function - called on keyboard events
	glutKeyboardFunc( keyboardFunc );
	// set the mouse function - called on mouse stuff
	glutMouseFunc( mouseFunc );

	g_window = NULL;

	g_pcm_left = new deque<MAD_SAMPLE>();
	g_pcm_right = new deque<MAD_SAMPLE>();

	g_fft_buffer_lock = new Mutex();

	StereoPCM out;
	out.left = g_pcm_left;
	out.right = g_pcm_right;
	out.sample_rate = 0;

	// We could start playing as we decode, but decode is really fast
	// Note it does mean the memory requirements for music are pretty huge,
	// if we want songs longer than like 10 minutes it would probably be wise
	// to do this on-demand
	decode(fildes, &out);

	close( fildes );

	g_sampleRate = out.sample_rate;

	// Once we detect the sample rate, set up the local energy buffer
	assert( g_sampleRate.value() > 0 );

	init_global_subband_widths();

	/*
	ProtoScript::Freeze( "test_script.ts", &out );
	delete ProtoScript::UnFreeze( "test_script.ts" );
	*/

	// create the RtAudio
	try {
		audio = new RtAudio();

		outParams.deviceId = audio->getDefaultOutputDevice();
		outParams.nChannels = NCHANNELS;

		audio->openStream(
				&outParams,
				NULL,
				RTAUDIO_FLOAT32,
				g_sampleRate.value(),
				&g_buffer_size,
				audioCallback );

		for( uint32_t i = 0; i < NSUBBANDS; i++ ) {
			g_subband_detector[i] =
				new BeatDetector( g_sampleRate.value() / g_buffer_size );
		}

		g_window = new RT_SAMPLE[g_buffer_size];
		g_fft_buffer = new RT_SAMPLE[g_buffer_size];

		hanning( g_window, g_buffer_size );

	} catch( RtError & err ) {
		cout << err.getMessage();
		if( audio ) {
			delete audio;
		}
		exit(1);
	}

	try {
		audio->startStream();
	} catch( RtError & err ) {
		cout << err.getMessage();
		goto cleanup;
	}

	gettimeofday( &g_last_draw, NULL );

	// let GLUT handle the current thread from here
	glutMainLoop();

	// something needs to happen here
	while( true ) {
		usleep( 10000 );
	}

	// if we get here, then stop!
	try {
		audio->stopStream();
	} catch( RtError & err ) {
		cout << err.getMessage();
	}

cleanup:
	audio->closeStream();
	delete audio;

	delete g_pcm_left;
	delete g_pcm_right;

	for( uint32_t sb = 0; sb < NSUBBANDS; sb++ ) {
		delete g_subband_detector[sb];
	}

	return 0;
}

//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( GLsizei w, GLsizei h )
{
	// save the new window size
	g_width = w; g_height = h;
	// map the view port to the client area
	glViewport( 0, 0, w, h );
	// set the matrix mode to project
	glMatrixMode( GL_PROJECTION );
	// load the identity matrix
	glLoadIdentity( );
	// create the viewing frustum
	gluPerspective( 45.0, (GLfloat) w / (GLfloat) h, 1.0, 300.0 );
	// set the matrix mode to modelview
	glMatrixMode( GL_MODELVIEW );
	// load the identity matrix
	glLoadIdentity( );
	// position the view point
	gluLookAt( 0.0f, 0.0f, 10.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f );
}

//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
	switch( key )
	{
		case 'Q':
		case 'q':
			exit(1);
			break;
	}

	glutPostRedisplay( );
}

//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y )
{
	if( button == GLUT_LEFT_BUTTON )
	{
		// when left mouse button is down
		if( state == GLUT_DOWN )
		{
		}
		else
		{
		}
	}
	else if ( button == GLUT_RIGHT_BUTTON )
	{
		// when right mouse button down
		if( state == GLUT_DOWN )
		{
		}
		else
		{
		}
	}
	else
	{
	}

	glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
	// render the scene
	glutPostRedisplay( );
}

void drawEnergy()
{
	// line width
	glLineWidth( 2.0 );

	glPushMatrix();

	glColor3f( 0.0, 1.0, 0.0 );

	glBegin( GL_LINE_STRIP );

	const GLfloat xStart = -5;
	GLfloat x = xStart;

	complex *cbuf = (complex *)g_fft_buffer;
	RT_SAMPLE my_buffer[g_buffer_size/2];

	g_fft_buffer_lock->Lock();
	{
		for( uint32_t i = 0; i < g_buffer_size/2; i++ ) {
			my_buffer[i] = cmp_abs( cbuf[i] );
		}
	}
	g_fft_buffer_lock->Unlock();

	uint32_t start = 0;
	for( uint32_t sb = 0; sb < NSUBBANDS; sb++ ) {

		GLfloat xInc = ::fabs( 2 * xStart / NSUBBANDS ) / g_subband_width[sb];

		for( uint32_t i = start; i < start + g_subband_width[sb]; i++ ) {
			glVertex2f( x, 9.0 * sqrt( my_buffer[i] ) );
			x += xInc;
		}

		start += g_subband_width[sb];
	}

	glEnd();

	glBegin( GL_LINE_STRIP );

	x = xStart;
	GLfloat xInc = ::fabs( 4 * xStart / g_buffer_size );
	for( uint32_t i = 0; i < g_buffer_size / 2; i++ ) {
		glVertex2f( x, 9.0 * sqrt( my_buffer[i] ) - 2.0 );
		x += xInc;
	}

	glEnd();

	glPopMatrix();
}

void drawBeat()
{
	glPushMatrix();

	const float fade_every_us = 10e6 / 800; /* 800 fade steps/second */
	const float fade_base = 0.95; /* 100 fades takes us to 0.5% of original val */

	static uint64_t us_since_last_fade = 0;
	struct timeval tv;
	gettimeofday( &tv, NULL );
	uint64_t elapsed_us =
		( tv.tv_sec - g_last_draw.tv_sec ) * 10e6 +
		( tv.tv_usec - g_last_draw.tv_usec );

	g_last_draw = tv;

	assert( elapsed_us >= 0 );

	us_since_last_fade += elapsed_us;

	// Handle differing frame rates by calculating a fade factor since the last
	//  update based on desired fade rate.

	// Fade factor depends on how much we need to fade since the last refresh
	// in order to fade to 0.5% of original color in 1/2 second
	uint32_t n_fades = us_since_last_fade / fade_every_us;
	float fade_factor = 1.0f;

	if( n_fades > 0 ) {
		us_since_last_fade = 0;
		fade_factor *= pow( fade_base, n_fades );

		for( uint32_t sb = 0; sb < NSUBBANDS; sb++ ) {
			for( uint32_t ci = 0; ci < 3; ci++ ) { /* RGB */
				g_subband_color[sb][ci] *= fade_factor;
			}
		}
	}

	for( uint32_t sb = 0; sb < NSUBBANDS; sb++ ) {
		if( g_subband_detector[sb]->is_beat() ) {
			g_subband_color[sb][0] = 1.0;
			g_subband_color[sb][1] = (GLfloat)(sb+1) / (GLfloat)NSUBBANDS;
			g_subband_color[sb][2] = 0.0;
		}
	}

	GLfloat x = -5;
	GLfloat xInc = ::fabs( 2 * x / NSUBBANDS );
	for( uint32_t sb = 0; sb < NSUBBANDS; sb++  ) {
		glColor3f( g_subband_color[sb][0],
				g_subband_color[sb][1],
				g_subband_color[sb][2] );
		glRectf( x, -5, x + xInc, 5 );
		x += xInc;
	}

	glPopMatrix();
}

//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
	// clear the color and depth buffers
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	drawBeat();

	drawEnergy();

	// flush!
	glFlush( );
	// swap the double buffer
	glutSwapBuffers( );
}
